-- |
-- Module      : Network.TLS.Extra
-- License     : BSD-style
-- Maintainer  : Vincent Hanquez <vincent@snarc.org>
-- Stability   : experimental
-- Portability : unknown
--
-- default values and ciphers
module Network.TLS.Extra
    ( module Network.TLS.Extra.Cipher
    , module Network.TLS.Extra.FFDHE
    ) where

import Network.TLS.Extra.Cipher
import Network.TLS.Extra.FFDHE
